// -*- C++ -*-
// $RCSfile: stringimage.C,v $
// $Revision: 1.10 $
// $Author: langer $
// $Date: 2004/03/18 00:18:49 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/IO/bitoverlay.h"
#include "common/IO/stringimage.h"
#include "common/trace.h"
#include "common/tostring.h"
#include <set>
#include <string.h>		// for memcpy

StringImage::StringImage(const ICoord *isighs, const Coord *sighs)
  : size_(*sighs), isize_(*isighs)
{
  int n = 3*isize_(0)*isize_(1);
  data = new unsigned char[n+1];
  data[n] = '\0';
}

StringImage::~StringImage() {
  delete [] data;
}

int StringImage::getOffset(const ICoord &where) const {
  return 3*((isize_(1)-where(1)-1)*isize_(0) + where(0));
}

void StringImage::set(const ICoord *pos, const CColor *color) {
  int offset = getOffset(*pos);
  data[offset] = (unsigned char)(255*color->getRed());
  data[offset+1] = (unsigned char)(255*color->getGreen());
  data[offset+2] = (unsigned char)(255*color->getBlue());
}

CColor StringImage::get(const ICoord *pos) const {
  int offset = getOffset(*pos);
  return CColor(data[offset]/255., data[offset+1]/255., data[offset+2]/255.);
}


ShapedStringImage *StringImage::blend(const BitmapOverlay *bitmap,
				      double opacity)
  const
{
  return new ShapedStringImage(*this, *bitmap, opacity);
}

ShapedStringImage *StringImage::mask(const BoolArray &bits) const {
  return new ShapedStringImage(*this, bits);
}


ShapedStringImage::ShapedStringImage(const StringImage &src,
				     const BitmapOverlay &bitmap,
				     double opacity)
  : StringImage(&src.sizeInPixels(), &src.size())
{
  CColor blend = bitmap.getFG(); // blend original image colors with this color
  double blendR = blend.getRed(); // 0 <= blendR <= 1
  double blendG = blend.getGreen();
  double blendB = blend.getBlue();
  const BoolArray &bdata = bitmap.data;
  for(BoolArray::const_iterator i=bdata.begin(); i!=bdata.end(); ++i) {
    if(*i) {
      int offset = getOffset(i.coord());
      unsigned char r = src.getString()[offset]; // 0 <= r <= 255
      unsigned char g = src.getString()[offset+1];
      unsigned char b = src.getString()[offset+2];
      data[offset] = (unsigned char) (blendR*(255*opacity + (1-opacity)*r));
      data[offset+1] = (unsigned char) (blendG*(255*opacity + (1-opacity)*g));
      data[offset+2] = (unsigned char) (blendB*(255*opacity + (1-opacity)*b));
    }
  }
  findBGcolor(bitmap.data);
  applyMask(bitmap.data);
}

ShapedStringImage::ShapedStringImage(const StringImage &src,
				     const BoolArray &bits)
  : StringImage(&src.sizeInPixels(), &src.size())
{
  (void) memcpy(data, src.getString(), isize_(0)*isize_(1));
  findBGcolor(bits);
  applyMask(bits);
}

struct ltCharColor {
  bool operator()(const CharColor &a, const CharColor &b) const {
    if(a.r < b.r) return true;
    if(a.r > b.r) return false;
    if(a.g < b.g) return true;
    if(a.g > b.g) return false;
    if(a.b < b.b) return true;
    return false;
  }
};

std::ostream &operator<<(std::ostream &os, const CharColor &cc) {
  os << "(" << (int) cc.r << ", " << (int) cc.g << ", " << (int) cc.b << ")";
  return os;
}

void ShapedStringImage::applyMask(const BoolArray &bits) {
  for(BoolArray::const_iterator i=bits.begin(); i!=bits.end(); ++i)
    if(!*i) {
      int offset = getOffset(i.coord());
      data[offset] = shapecolor_.r;
      data[offset+1] = shapecolor_.g;
      data[offset+2] = shapecolor_.b;
    }
}

void ShapedStringImage::findBGcolor(const BoolArray &bits) {
  // Find a color that's not in the image.
  std::set<CharColor, ltCharColor> palette;
  for(BoolArray::const_iterator i=bits.begin(); i!=bits.end(); ++i) {
    if(*i) {
      int offset = getOffset(i.coord());
      (void) palette.insert(CharColor(data+offset));
//       std::cerr << "ShapedStringImage::findBGcolor: " << offset << " " << CharColor(data+offset) << std::endl;
    }
  }
  for(int r=0; r<256; r++)
    for(int g=0; g<256; g++)
      for(int b=0; b<256; b++) {
	CharColor testcolor(r,g,b);
	if(palette.count(testcolor) == 0) {
	  shapecolor_= testcolor;
	  return;
	}
      }
  throw ErrProgrammingError("Failed to find suitable background color",
			    __FILE__, __LINE__);
}

CColor ShapedStringImage::getBG() const {
//   std::cerr << "ShapedStringImage::getBG: " << shapecolor_ << std::endl;
  return shapecolor_.ccolor();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//


void inline hexbyte(char *&dest,int num)
{
  dest[0]=num/16;
  dest[1]=num%16;

  if(dest[0]>9)
    dest[0]+='A'-10;
  else
    dest[0]+='0';

  if(dest[1]>9)
    dest[1]+='A'-10;
  else
    dest[1]+='0';
  dest+=2;
}

//gives the string in 24 bit hexadecimal format. Used for post script output.
std::string StringImage::hexstringimage() const
{
  int bytenum=isize_(1)*isize_(0)*3;
  int stringlen=bytenum*2+(bytenum/24);
  char *dest=new char[stringlen+1];
  char *dpos=dest;
  unsigned char *spos=data;

  for(int i=0; i<bytenum; i++) {
    *dpos=*spos/16;		// high order hex digit

    if(*dpos>9)
      *dpos+='a'-10;
    else
      *dpos+='0';
    dpos++;

    *dpos=*spos%16;		// low order hex digit
    if(*dpos>9)
      *dpos+='a'-10;
    else
      *dpos+='0';
    dpos++;

    spos++;
    if(i%24==23)
      *dpos++='\n';
  }
  *dpos=0;

  std::string ret(dest,stringlen);
  delete[] dest;
  return ret;
}

//gives the string in 24 bit hexadecimal format. Used for post script output.
std::string StringImage::hexstringmask(CColor *shape_color) const
{
  int bytenum=isize_(1)*(isize_(0)/8+int(isize_(0)%8!=0));
  int stringlen=bytenum*2+(bytenum/24);
  char *dest=new char[stringlen+1];
  char *dpos=dest;
  unsigned char *spos=data;
  unsigned long color = int(255*shape_color->getRed()) |
    (int(255*shape_color->getGreen())<<8) |
    (int(255*shape_color->getBlue())<<16);
  int num;
  int pos=0;
  int i,j;

  for(i=0; i<isize_(1); i++)
  {
    num=0;
    for(j=0; j<isize_(0); j++)
    {
      num<<=1;
      num|=((*(unsigned long *)spos&0xFFFFFF)==color);
      spos+=3;
      if(j%8==7)
      {
	hexbyte(dpos,num);
	if(pos++%24==23)
	  *dpos++='\n';
	num=0;
      }
    }
    if(j%8!=0)
    {
      num<<=8-j%8;
      hexbyte(dpos,num);
      if(pos++%24==23)
	*dpos++='\n';
    }
  }
  *dpos=0;

  std::string ret(dest,stringlen);
  delete[] dest;
  return ret;
}
